import numpy as np
from scipy.interpolate import interpn
import matplotlib.pyplot as plt
from skimage.transform.integral import integral_image
from scipy.ndimage.filters import gaussian_filter1d
from .config import (
    GAUSSIAN_RELATIVE_SIGMA,
    CORNER_DETECTION_MAX_PEAKS,
    CORNER_DETECTION_PEAK_THRESH,
)
from .summary import Summary
from scipy.signal import find_peaks

summary = Summary("locate_corners")


def inner_corners_model(n_cols, m_rows):
    x, y = np.meshgrid(np.arange(1, n_cols), np.arange(1, m_rows))
    return np.array([x.flatten(), y.flatten()], dtype=np.float64).T


def outer_corners_model(n_cols, m_rows):
    x = (
        list(range(0, n_cols + 1))
        + [n_cols] * (m_rows - 1)
        + list(reversed(range(0, n_cols + 1)))
        + [0] * (m_rows - 1)
    )
    y = (
        [0] * (n_cols + 1)
        + list(range(1, m_rows))
        + [m_rows] * (n_cols + 1)
        + list(reversed(range(1, m_rows)))
    )

    return np.array([x, y], dtype=np.float64).T


def _corner_patches(coords, patch_size):
    d = patch_size / 2
    coords = np.repeat(np.expand_dims(coords, 1), 4, 1)
    coords[:, 0] -= d
    coords[:, 1, 0] += d
    coords[:, 1, 1] -= d
    coords[:, 2] += d
    coords[:, 3, 0] -= d
    coords[:, 3, 1] += d
    return coords


def _extract_profile2d(img, patch, transform, n_samples):
    n_samples = 80
    x, y = np.meshgrid(
        np.linspace(patch[0, 0], patch[1, 0], n_samples),
        np.linspace(patch[0, 1], patch[3, 1], n_samples),
    )
    xy = np.array([x.flatten(), y.flatten()]).T
    xy_t = transform(xy)

    # simple projection to image bounds
    xy_t[xy_t[:, 0] < 0.0, 0] = 3
    xy_t[xy_t[:, 0] > img.shape[1] - 1, 0] = img.shape[1] - 3
    xy_t[xy_t[:, 1] < 0.0, 1] = 3  # 0 gives artifacts
    xy_t[xy_t[:, 1] > img.shape[0] - 1, 1] = img.shape[0] - 3

    # warp patch
    yx_t = np.flip(xy_t, 1)
    v = interpn(
        (np.arange(img.shape[0]), np.arange(img.shape[1])),
        img,
        yx_t,
        method="linear",
        bounds_error=False,
        fill_value=0.0,
    )
    summary.put(
        "patch_{:d}_{:d}".format(int(patch[0, 0] + 1), int(patch[0, 1] + 1)),
        v.reshape(n_samples, n_samples),
    )
    return x, y, v.reshape(n_samples, n_samples)


def _find_stops(img, dim, sigma, patch, edgetype="ridge"):
    img = img.T if dim == 0 else img

    # calculate downsampling
    size = 0.5 * np.sum(img.shape)

    # extract profile of cumsum along di
    profile = np.sum(img, 1)
    profile_smooth = gaussian_filter1d(profile, sigma)
    profile_smooth = profile_smooth - np.min(profile_smooth)
    profile_smooth = profile_smooth / (np.max(profile_smooth) + 1e-5)

    # calculate gradient of that
    grad_smooth = np.gradient(profile_smooth)
    grad2_smooth = np.gradient(grad_smooth)

    summary.put(
        "patch_{:d}_{:d}_profile_{}".format(
            int(patch[0, 0] + 1), int(patch[0, 1] + 1), "x" if dim == 0 else "y"
        ),
        profile,
    )
    summary.put(
        "patch_{:d}_{:d}_profile_grad_{}".format(
            int(patch[0, 0] + 1), int(patch[0, 1] + 1), "x" if dim == 0 else "y"
        ),
        grad_smooth,
    )

    # find peaks in first and second derivative of profile
    min_distance = img.shape[0] // CORNER_DETECTION_MAX_PEAKS
    peaks, _ = find_peaks(
        grad2_smooth,
        distance=min_distance,
        height=CORNER_DETECTION_PEAK_THRESH * np.mean(np.abs(grad2_smooth)),
    )
    maxpeaks, _ = find_peaks(
        grad_smooth,
        distance=min_distance,
        height=CORNER_DETECTION_PEAK_THRESH * np.mean(np.abs(grad_smooth)),
    )
    minpeaks, _ = find_peaks(
        -grad_smooth,
        distance=min_distance,
        height=CORNER_DETECTION_PEAK_THRESH * np.mean(np.abs(grad_smooth)),
    )

    # plt.subplot(1,2,1)
    # plt.plot(np.arange(profile.shape[0]), grad_smooth)
    # plt.scatter(peaks, grad_smooth[peaks])
    # plt.subplot(1,2,2)
    # plt.plot(np.arange(profile.shape[0]), grad2_smooth)
    # plt.scatter(peaks, grad2_smooth[peaks])
    # plt.show()

    # threshold for finding minimum/maximum corresponding to peak in 2nd derivative
    eps = img.shape[0] // 10

    if edgetype in ("up", "down"):
        # filter all peaks in 2nd derivative that are not close to a maximum/minimum
        compare = maxpeaks if edgetype == "up" else minpeaks
        peaks_new = []
        for peak in peaks.tolist():
            if np.any(np.abs(peak - compare) <= eps):
                peaks_new.append(peak)

        # if multiple found, return the one closest to the center
        if len(peaks_new) > 1:
            center = img.shape[0] // 2
            distances = np.abs(np.array(peaks_new) - center)
            return peaks_new[np.argmin(distances)]
        elif len(peaks_new) == 1:
            return peaks_new[0]
        else:
            return None

    else:
        # filter all peaks in 2nd derivative that are not close to a maximum and a minimum
        peaks_new = []
        for peak in peaks.tolist():
            if np.any(np.abs(peak - maxpeaks) <= eps) and np.any(
                np.abs(peak - minpeaks) <= eps
            ):
                peaks_new.append(peak)

        # if number of remaining peaks uneven, return the middle one
        if len(peaks_new) % 2 == 1:
            return peaks_new[len(peaks_new) // 2]
        # otherwise take the one closest to the center
        elif len(peaks_new) > 0:
            center = img.shape[0] // 2
            distances = np.abs(np.array(peaks_new) - center)
            return peaks_new[np.argmin(distances)]
        else:
            return None


def _determine_sampling_and_filter(patch_t):
    patch_width = max(
        np.linalg.norm(patch_t[1] - patch_t[0]), np.linalg.norm(patch_t[3] - patch_t[2])
    )
    patch_height = max(
        np.linalg.norm(patch_t[2] - patch_t[1]), np.linalg.norm(patch_t[3] - patch_t[0])
    )
    n_samples = max(patch_height, patch_width)

    sigma = GAUSSIAN_RELATIVE_SIGMA * n_samples
    s = np.floor(2 / 3 * np.pi * sigma)

    if s > 1:
        sigma_s = (3 * s) / (2 * np.pi)
        if sigma ** 2 - sigma_s ** 2 < 0.5 ** 2:
            # convince Nyquist
            s -= 1
            sigma_s = (3 * s) / (2 * np.pi)
        sigma_filter = np.sqrt(sigma ** 2 - sigma_s ** 2)
        n_samples /= s
    else:
        sigma_filter = sigma

    return int(n_samples), sigma_filter


def fit_inner_corners(img, coords, transform, patch_size):
    patches = _corner_patches(coords, patch_size)
    accepted_flags = list()
    corners = list()

    # determine #samples and filter for one example only
    n_samples, sigma_filter = _determine_sampling_and_filter(
        transform(patches[len(patches) // 2])
    )

    for i in range(patches.shape[0]):
        _, _, v = _extract_profile2d(img, patches[i], transform, n_samples)
        # plt.imshow(v, cmap='gray')
        # plt.show()

        # find stop in x and y direction
        xc = _find_stops(v, 0, sigma_filter, patches[i])
        yc = _find_stops(v, 1, sigma_filter, patches[i])
        # plt.imshow(v, cmap='gray')
        # plt.scatter(xc, yc)
        # plt.show()

        # plt.imshow(img, cmap='gray')
        # xcc = (xc/v.shape[1]) - 0.5
        # ycc = (yc/v.shape[0]) - 0.5
        # xcc *= patch_size
        # ycc *= patch_size
        # ct = transform(coords[i]+np.array([xcc, ycc]))
        # plt.scatter(ct[0], ct[1], s=1)
        # plt.show()

        # if both succeded
        if xc is not None and yc is not None:

            # xstops and ystops is given in patch coordinates -> transform to image coordinates
            x_size, y_size = v.shape[1], v.shape[0]
            corner = [xc / x_size, yc / y_size]
            accepted_flags.append(True)
            corners.append(corner)
        else:
            accepted_flags.append(False)
            corners.append([0, 0])

        # plt.imshow(v, cmap = plt.get_cmap('gray'))

        # if accepted:
        #    plt.plot([corner[0]*x_size], [corner[1]*y_size], marker = 'x', color = 'green')
        # else:
        #    plt.plot([corner[0]*x_size], [corner[1]*y_size], marker = 'x', color = 'red')

        # plt.show()
        # plt.clf()

    corners = np.array(corners)
    corners -= 0.5
    corners *= patch_size
    return coords + corners, np.array(accepted_flags)


def fit_outer_corners(
    img, coords, transform, patch_size, n_cols, m_rows, is_module_detail
):

    if is_module_detail:
        # determine, on which sides the module is continued
        x_profile = np.sum(img, 1)
        y_profile = np.sum(img, 0)
        x_thresh = 0.3 * (np.max(x_profile) - np.min(x_profile)) + np.min(x_profile)
        y_thresh = 0.3 * (np.max(y_profile) - np.min(y_profile)) + np.min(y_profile)
        cont_left = np.median(y_profile[:10]) > y_thresh
        cont_right = np.median(y_profile[-10:]) > y_thresh
        cont_top = np.median(x_profile[:10]) > x_thresh
        cont_bottom = np.median(x_profile[-10:]) > x_thresh
        # plt.plot(np.arange(x_profile.shape[0]), x_profile)
        # plt.show()
        # plt.plot(np.arange(y_profile.shape[0]), y_profile)
        # plt.show()
        # print(x_thresh, y_thresh)
    else:
        cont_left, cont_right, cont_top, cont_bottom = False, False, False, False

    # depending on the orientation of the model, we need to assign differently
    x0 = transform(np.array([[0, 0]])).flatten()
    if np.linalg.norm(x0 - np.array([0, 0])) < np.linalg.norm(
        x0 - np.array([img.shape[0], 0])
    ):
        # horizontal orientation
        x_cont_left = cont_left
        x_cont_right = cont_right
        y_cont_left = cont_top
        y_cont_right = cont_bottom
    else:
        x_cont_left = cont_top
        x_cont_right = cont_bottom
        y_cont_left = cont_right
        y_cont_right = cont_left

    patches = _corner_patches(coords, patch_size)
    accepted_flags = list()
    corners = list()

    # determine #samples and filter for one example only
    n_samples, sigma_filter = _determine_sampling_and_filter(
        transform(patches[len(patches) // 2])
    )

    for i in range(patches.shape[0]):
        _, _, v = _extract_profile2d(img, patches[i], transform, n_samples)
        # plt.imshow(v, cmap='gray')
        # plt.show()

        if i <= n_cols and not y_cont_left:
            y_edgetype = "up"
        elif i >= n_cols + m_rows and i <= 2 * n_cols + m_rows and not y_cont_right:
            y_edgetype = "down"
        else:
            y_edgetype = "ridge"

        if i >= n_cols and i <= n_cols + m_rows and not x_cont_right:
            x_edgetype = "down"
        elif (i >= 2 * n_cols + m_rows or i == 0) and not x_cont_left:
            x_edgetype = "up"
        else:
            x_edgetype = "ridge"

        xc = _find_stops(v, 0, sigma_filter, patches[i], x_edgetype)
        yc = _find_stops(v, 1, sigma_filter, patches[i], y_edgetype)

        if xc is not None and yc is not None:
            x_size, y_size = v.shape[1], v.shape[0]
            xc /= x_size
            yc /= y_size
            corner = [xc, yc]
            accepted_flags.append(True)
            corners.append(corner)
        else:
            accepted_flags.append(False)
            corners.append([0, 0])

        # plt.imshow(v, cmap = plt.get_cmap('gray'))

        # if accepted:
        #    plt.plot([corner[0]*x_size], [corner[1]*y_size], marker = 'x', color = 'green')
        # else:
        #    plt.plot([corner[0]*x_size], [corner[1]*y_size], marker = 'x', color = 'red')

        # plt.show()
        # plt.clf()

    corners = np.array(corners)
    corners -= 0.5
    corners *= patch_size
    return coords + corners, np.array(accepted_flags)



from pvinspect.data import datasets
from pvinspect.preproc.detection import (
    locate_multiple_modules,
    locate_module_and_cells,
    segment_modules,
    segment_cells,
)
from pvinspect.data.image import *
from test.utilities import *


def test_multi_segmentation():
    # load images
    _, imgs = datasets.multi_module_detection(N=2)

    # perform multi module segmentation
    modules = locate_multiple_modules(imgs, rows=6, cols=10)

    # check result
    assert len(modules) == 12
    assert isinstance(modules[0].get_meta("multimodule_original"), Image)
    assert modules[0].get_meta("multimodule_original").path == imgs[0].path

    # check that plotting does not fail
    modules[0].get_meta("multimodule_original").show()

    # perform precise module localization and segmentation
    modules = locate_module_and_cells(modules)
    modules_crop = segment_modules(modules)

    # check result
    assert len(modules) == 12
    assert_equal(
        modules_crop[0].get_meta("transform")(np.array([[0.0, 0.0]])),
        np.array([0.0, 0.0]),
    )

    # make sure that original is preserved
    assert isinstance(modules_crop[0].get_meta("multimodule_original"), Image)
    assert modules_crop[0].get_meta("multimodule_original").path == imgs[0].path

    # check show
    modules_crop.head()

    # check segmentation into cells
    cells = segment_cells(modules_crop[0])
    assert len(cells) == 60

    # make sure that original is preserved
    assert isinstance(cells[0].get_meta("multimodule_original"), Image)
    assert cells[0].get_meta("multimodule_original").path == imgs[0].path


def test_single_segmentation():
    # load images
    img = datasets.poly10x6(N=1)[0]

    # perform detection
    module = locate_module_and_cells(img)

    # check result
    assert module.has_meta("transform")

    # check that show result does not fail
    module.show()

    # perform segmentation into cells
    cells = segment_cells(module)

    # check
    assert len(cells) == 60



from test.utilities import assert_equal

import cv2
import numpy as np
from pvinspect.data import datasets
from pvinspect.data.image import *
from pvinspect.data.io import save_image
from pvinspect.preproc.calibration import *
from pvinspect.preproc.calibration import _calibrate_flatfield, _compensate_flatfield
from pvinspect.preproc.stitching import *
from skimage.exposure import rescale_intensity
from skimage.metrics import structural_similarity


def _prepare_ff_data():
    ds = datasets.calibration_ipv40CCD_FF(N=2)
    del ds["1A"]
    imgs_list = []
    ex_list = []
    ex_all = [int(x[0]) for x in ds.keys()]
    ex_max = max(ex_all)
    for ex in ex_all:
        for img in ds["{:d}A".format(ex)]:
            imgs_list.append(img)
            ex_list.append(ex / ex_max)

    return ImageSequence(imgs_list, same_camera=True), ex_list


def _prepare_ff_test_img():
    ds = datasets.calibration_ipv40CCD_FF(N=2)
    del ds["1A"]
    imgs_list = []
    ex_list = []
    ex_all = [int(x[0]) for x in ds.keys()]
    ex_max = max(ex_all)
    for ex in ex_all:
        for img in ds["{:d}A".format(ex)]:
            imgs_list.append(img)
            ex_list.append(ex / ex_max)
    test_img = ModuleImage(
        ds["{:d}A".format(ex_all[1])][0]._data, EL_IMAGE, path=Path() / "test.png"
    )
    return test_img


def _make_image_seq(imgs):
    imgs = [ModuleImage(img, EL_IMAGE, None) for img in imgs]
    return ModuleImageSequence(imgs, True, False)


def _prepare_stitching_test_img():
    image = datasets.poly10x6(1)[0].data.T[:, :1930]
    height = image.shape[0]
    width = image.shape[1]

    img0 = Image(image[600:, 0 : int(width * 2 // 3)])
    img1 = Image(image[600:, int(width // 3) :])
    return img0, img1


def test_calibrate_flatfield_linear():
    img0 = np.random.random((10, 10)) / 10
    img1 = (-np.random.random((10, 10)) / 10) + 1.0
    coeff = 1 / (img1 - img0)
    coeff *= np.mean(img1)

    coeff_result = _calibrate_flatfield([img0, img1], [0, 1.0], order=1)
    assert_equal(coeff_result[1], coeff)
    assert_equal(coeff_result[0], -coeff * img0)


def test_calibrate_flatfield_quadratic():
    img0 = np.random.random((10, 10)) / 10
    img1 = np.random.random((10, 10)) / 10 + 0.5 - 0.05
    img2 = (-np.random.random((10, 10)) / 10) + 1.0
    mean = np.mean(img2)

    def compensate(img, coeff):
        return coeff[2] * img ** 2 + coeff[1] * img + coeff[0]

    coeff = _calibrate_flatfield([img0, img1, img2], [0, 0.5, 1.0], order=2)
    assert_equal(np.zeros_like(img0), compensate(img0, coeff))
    assert_equal(np.full_like(img1, 0.5 * mean), compensate(img1, coeff))
    assert_equal(np.full_like(img1, 1.0 * mean), compensate(img2, coeff))


def test_calibrate_flatfield_least_squares_linear():
    img0 = np.random.random((10, 10)) / 10
    img2 = (-np.random.random((10, 10)) / 10) + 1.0
    coeff = 1 / (img2 - img0)
    img1 = (np.full_like(img0, 0.5) + img0) / coeff
    coeff *= np.mean(img2)
    coeff_result = _calibrate_flatfield([img0, img1, img2], [0, 0.5, 1.0], order=1)

    assert_equal(coeff_result[1], coeff, 0.1)
    assert_equal(coeff_result[0], -coeff * img0, 0.1)


def test_compensate_flatfield():
    img = np.random.random((10, 10))
    coeff = np.random.random((4, 10, 10)) / 10
    target = coeff[0] + coeff[1] * img + coeff[2] * img ** 2 + coeff[3] * img ** 3

    res = _compensate_flatfield(_make_image_seq([img]), coeff)[0].data
    assert_equal(res, target, 0.1)


def test_calibrate_and_compensate_flatfield():
    img0 = np.random.random((10, 10)) / 10
    img1 = np.random.random((10, 10)) / 10 + 0.5 - 0.05
    img2 = (-np.random.random((10, 10)) / 10) + 1.0
    mean = np.mean(img2)

    coeff = _calibrate_flatfield([img0, img1, img2], [0, 0.5, 1.0], order=2)
    compensated = _compensate_flatfield(_make_image_seq([img0, img1, img2]), coeff)
    assert_equal(np.zeros_like(img0), compensated[0].data)
    assert_equal(np.full_like(img1, 0.5 * mean), compensated[1].data)
    assert_equal(np.full_like(img1, 1.0 * mean), compensated[2].data)


def test_calibrate_and_compensate_uint_preserves_range():
    img0 = np.array([[0]], dtype=DTYPE_UNSIGNED_INT)
    img1 = np.array([[10000]], dtype=DTYPE_UNSIGNED_INT)
    img_test = np.array([[8000]], dtype=DTYPE_UNSIGNED_INT)

    coeff = calibrate_flatfield(_make_image_seq([img0, img1]), [0, 1.0])
    compensated = compensate_flatfield(_make_image_seq([img_test]), coeff)
    assert_equal(compensated[0].data, 8000)


def test_calibrate_and_compensate_float_preserves_range():
    img0 = np.array([[0.0]], dtype=DTYPE_FLOAT)
    img1 = np.array([[1.0]], dtype=DTYPE_FLOAT)
    img_test = np.array([[0.8]], dtype=DTYPE_FLOAT)

    coeff = calibrate_flatfield(_make_image_seq([img0, img1]), [0, 1.0])
    compensated = compensate_flatfield(_make_image_seq([img_test]), coeff)
    assert_equal(compensated[0].data, 0.8)


def test_calibrate_and_compensate_int_preserves_range():
    img0 = np.array([[-10]], dtype=DTYPE_INT)
    img1 = np.array([[10]], dtype=DTYPE_INT)
    img_test = np.array([[0]], dtype=DTYPE_INT)

    coeff = calibrate_flatfield(_make_image_seq([img0, img1]), [0, 1.0])
    compensated = compensate_flatfield(_make_image_seq([img_test]), coeff)
    assert_equal(compensated[0].data, 0)


def test_calibrate_and_compensate_flatfield_realdata():
    imgs_list, ex_list = _prepare_ff_data()
    test_img = _prepare_ff_test_img()
    coeff = calibrate_flatfield(imgs_list, ex_list)
    compensated = compensate_flatfield(test_img, coeff)
    data = compensated.data
    assert data.std() <= 0.01 * np.mean(test_img.data)


def test_calibrate_distortion():
    ds = datasets.calibration_ipv40CCD_distortion(N=3)
    res = calibrate_distortion(ds, (7, 7))

    assert len(res) == 4
    assert res[0].shape == (3, 3)
    assert res[1].shape == (1, 5)
    assert res[2].shape == (3, 3)
    assert len(res[3]) == 4


def test_compensate_distortion_identity():
    imgs = datasets.calibration_ipv40CCD_distortion(N=2)
    mtx = np.identity(3)
    dist = np.zeros((1, 5), dtype=np.float32)
    roi = (0, 0, imgs.shape[0], imgs.shape[1])
    imgs_comp = compensate_distortion(imgs, mtx, dist, mtx, roi)

    assert (imgs_comp[0].data - imgs[0].data).sum() == 0
    assert (imgs_comp[1].data - imgs[1].data).sum() == 0


def test_calibrate_compensate_distortion_realdata():
    ds = datasets.calibration_ipv40CCD_distortion()
    res = calibrate_distortion(ds, (7, 7))
    ds_comp = compensate_distortion(ds, *res)

    for img, comp in zip(ds, ds_comp):
        assert_equal(comp.data.mean(), img.data.mean(), 0.01 * img.data.max())


def test_calibration_object(tmp_path: Path):
    ff_imgs, ex_list = _prepare_ff_data()
    test_img = _prepare_ff_test_img()
    dist_imgs = datasets.calibration_ipv40CCD_distortion(N=8)

    # perform calibration
    calib = Calibration()
    calib.calibrate_flatfield(images=ff_imgs, targets=ex_list)
    calib.calibrate_distortion(images=dist_imgs, checkerboard_size=(7, 7))

    # save to disk
    calib.save(tmp_path / "calib.pck")

    # load from disk
    calib_loaded = load_calibration(tmp_path / "calib.pck")

    # perform compensation with both
    comp = calib.process(test_img)
    comp_loaded = calib_loaded.process(test_img)

    # assert both result in the same image
    assert_equal(comp.data, comp_loaded.data)


def test_flatfield_sequences_input():
    seqs = list(datasets.calibration_ipv40CCD_FF(N=2).values())
    seqs = [seqs[0], seqs[2]]
    test_img = _prepare_ff_test_img()

    coeff = calibrate_flatfield(seqs, targets=[0.0, 1.0])
    comp = compensate_flatfield(test_img, coeff)

    data = comp.data
    assert data.std() <= 0.01 * np.mean(seqs[1][0].data)


def test_stitching():
    (image_0, image_1,) = _prepare_stitching_test_img()
    images = (image_0, image_1)

    # test stitching and check size of result
    stitched = locate_and_stitch_modules(
        ImageSequence(images, True), 6, 6, 2, overlap_horizontal=2
    )
    assert_equal(stitched.shape[1] / stitched.shape[0], 10 / 6, precision=0.05)



import numpy as np
from copy import copy, deepcopy
import scipy.optimize
import scipy.interpolate
from abc import ABC, abstractmethod, abstractproperty
from functools import reduce
import cv2
from . import util


def _brown(coords, coeffs):
    N = len(coeffs)

    radius = np.sqrt(np.sum(coords ** 2, axis=1))
    factor = 1 + np.expand_dims(
        np.sum([kn * (radius ** (2 * (n + 1))) for n, kn in enumerate(coeffs)], axis=0),
        axis=-1,
    )
    coords = coords * factor

    return coords


def _brown_oder1_inverse(coords, k):

    if k == 0.0:
        return coords

    ru = np.sqrt(np.sum(coords ** 2, axis=1))
    D = (27 * k ** 2 * ru) ** 2 + 4 * (3 * k) ** 3

    rd = np.empty_like(D)
    g1 = D >= 0
    if np.any(g1):
        tmp = np.sqrt((ru[g1] ** 2) / 4 + 1 / (27 * k))
        rd[g1] = np.cbrt(1 / k * (ru[g1] / 2 + tmp)) + np.cbrt(
            1 / k * (ru[g1] / 2 - tmp)
        )
    g2 = D < 0
    if np.any(g2):
        rd[g2] = (
            2
            * np.sqrt(-3 * k)
            / (3 * k)
            * np.cos(
                1 / 3 * np.arccos(27 * k ** 2 * ru[g2] / (2 * np.sqrt(-((3 * k) ** 3))))
                + 4 * np.pi / 3
            )
        )

    factor = rd / ru
    coords = coords * factor.reshape(-1, 1)

    return coords


def distort(coords, params):
    return _brown(coords, params)


def distort_inverse(coords, params):
    if len(params) > 1:
        raise NotImplementedError(
            "Inversion not implemented for more than 1 coefficient"
        )
    return _brown_oder1_inverse(coords, params[0])


def _normalize_transform(coords, new_centroid=(0, 0), new_mean_distance=np.sqrt(2)):
    centroid = np.mean(coords, axis=0)
    scaling = new_mean_distance / np.mean(
        np.linalg.norm(coords - centroid.reshape(1, -1), axis=1)
    )
    return np.array([[scaling, 0, -centroid[0]], [0, scaling, -centroid[1]], [0, 0, 1]])


def find_homography(src, dest, compute_residual=False):
    """
        Estimate homography using normalized DLT

        TODO How handle cases where H[2,2] == 0?
    """
    N = src.shape[0]
    dt = src.dtype
    normalize = N > 4

    if normalize:
        # compute transforms that normalize coordinates such that
        # centroid is at origin with average distance of sqrt(2)
        src_T = _normalize_transform(src)
        dest_T = _normalize_transform(dest)

        # normalize
        src_n = apply_homography(src, src_T)
        dest_n = apply_homography(dest, dest_T)
    else:
        src_n = src
        dest_n = dest

    # DLT
    A = np.zeros((2 * N, 9), dtype=np.float64)
    A[:N, 0] = src_n[:, 0]
    A[:N, 1] = src_n[:, 1]
    A[:N, 2] = 1
    A[:N, 6] = -dest_n[:, 0] * src_n[:, 0]
    A[:N, 7] = -dest_n[:, 0] * src_n[:, 1]
    A[:N, 8] = -dest_n[:, 0]
    A[N:, 3] = -src_n[:, 0]
    A[N:, 4] = -src_n[:, 1]
    A[N:, 5] = -1
    A[N:, 6] = dest_n[:, 1] * src_n[:, 0]
    A[N:, 7] = dest_n[:, 1] * src_n[:, 1]
    A[N:, 8] = dest_n[:, 1]

    U, sigma, VT = np.linalg.svd(A)
    # residual = sigma[-1]
    H = VT[-1].reshape(3, 3)

    # denormalize
    if normalize:
        H = np.linalg.inv(dest_T).dot(H).dot(src_T)

    if np.abs(H[2, 2]) != 0.0:
        H /= H[2, 2]
    else:
        H = None

    if compute_residual and normalize:
        if H is not None:
            tmp = apply_homography(src, H)
            residual = np.mean(np.linalg.norm(tmp - dest, ord=2, axis=1))
            return H, residual
        else:
            return None, None
    elif compute_residual:
        return H, sigma[-1]
    else:
        return H


def apply_homography(src, H):
    assert src.shape[-1] == 2
    src_flat = src.reshape(-1, 2)
    res = H[:, :2].dot(src_flat.T)
    res += H[:, 2].reshape(3, 1)
    with np.errstate(divide="ignore"):
        res[:2] /= res[2]
    return res[:2].T.reshape(src.shape)


def apply_pinhole(src, Rt, A, d, inverse=False):
    if not inverse:
        Rt = Rt[:, [0, 1, 3]]
        src = apply_homography(src, Rt)
        src = distort(src, d)
        return apply_homography(src, A)
    else:
        Rt = np.linalg.pinv(Rt[:, [0, 1, 3]])
        A = np.linalg.pinv(A)
        src = apply_homography(src, A)
        src = distort_inverse(src, d)
        return apply_homography(src, Rt)


def find_homography_ransac(src, dest, error_thresh_px, compute_residual=False):
    def model_f(mask):
        return find_homography(src[mask], dest[mask])

    def err_f(model):
        src_t = apply_homography(src, model)
        return np.linalg.norm(src_t - dest, axis=1) < error_thresh_px

    return util.ransac(src.shape[0], 4, model_f, err_f)


def _solve_PnP_wrap(src, dest, A, dist, n_dist_coeff):
    # coords in OpenCV compatible format
    src2 = []
    for src_i in src:
        src_i = np.concatenate([src_i, np.zeros((src_i.shape[0], 1))], axis=1)
        src2.append(src_i.astype(np.float32))
    dest2 = [x.astype(np.float32) for x in dest]

    Rt = list()
    if n_dist_coeff <= 4:
        k = np.zeros(4)
    elif n_dist_coeff <= 5:
        k = np.zeros(5)
    elif n_dist_coeff <= 8:
        k = np.zeros(8)
    k[:n_dist_coeff] = dist

    for s, d in zip(src2, dest2):
        retval, rvec, tvec = cv2.solvePnP(s, d, A, k)
        R = util.rodrigues2matrix(rvec)
        Rt.append(np.concatenate([R, tvec.reshape(3, 1)], axis=1))

    return Rt


def _calibrate_camera_wrap(
    src, dest, image_width, image_height, n_dist_coeff, mask=None
):
    # coords in OpenCV compatible format
    src2 = []
    for src_i in src:
        src_i = np.concatenate([src_i, np.zeros((src_i.shape[0], 1))], axis=1)
        src2.append(src_i.astype(np.float32))
    dest2 = [x.astype(np.float32) for x in dest]

    # mask
    if mask is not None:
        src2_m, dest2_m = util.filter_lists([src2, dest2], mask)
    else:
        src2_m, dest2_m = src2, dest2

    # calibrate
    if n_dist_coeff > 4 or n_dist_coeff == 3:
        raise NotImplementedError()
    else:
        flags = cv2.CALIB_FIX_K3
        if n_dist_coeff < 3:
            flags += cv2.CALIB_ZERO_TANGENT_DIST
        if n_dist_coeff < 2:
            flags += cv2.CALIB_FIX_K2
        if n_dist_coeff == 0:
            flags += cv2.CALIB_FIX_K1
    _, A, k, rvecs, tvecs = cv2.calibrateCamera(
        objectPoints=src2_m,
        imagePoints=dest2_m,
        imageSize=(image_width, image_height),
        distCoeffs=np.zeros(6),
        cameraMatrix=np.zeros((3, 3)),
        flags=flags,
    )

    Rt = []
    if mask is not None:
        for s, d in zip(src2, dest2):
            retval, rvec, tvec = cv2.solvePnP(s, d, A, k)
            R = util.rodrigues2matrix(rvec)
            Rt.append(np.concatenate([R, tvec.reshape(3, 1)], axis=1))
    else:
        for rvec, tvec in zip(rvecs, tvecs):
            R = util.rodrigues2matrix(rvec)
            Rt.append(np.concatenate([R, tvec.reshape(3, 1)], axis=1))

    if n_dist_coeff > 0:
        return A, Rt, k.flatten()[:n_dist_coeff].tolist()
    else:
        return A, Rt


def find_full_model(src, dest, n_dist_coeff, image_width, image_height, ransac=False):
    N_images = len(src)

    # initial homographies
    H = [find_homography(s, d) for s, d in zip(src, dest)]
    errs = [
        np.mean(
            np.linalg.norm(
                src_i - apply_homography(dest_i, np.linalg.pinv(H_i)), axis=1
            )
        )
        for H_i, dest_i, src_i in zip(H, dest, src)
    ]

    # init intrinsics and extrinsics
    if ransac:
        thresh = 1.1 * np.array(errs)  # reprojection error must not become worse..

        def model_f(mask):
            return _calibrate_camera_wrap(src, dest, image_width, image_height, 0, mask)

        def err_f(model):
            A, Rt = model[0], model[1]
            dest_t = [
                apply_pinhole(dest_i, Rt_i, A, [0], inverse=True)
                for dest_i, Rt_i in zip(dest, Rt)
            ]
            err = [
                np.mean(np.linalg.norm(src_i - dest_t_i, axis=1))
                for src_i, dest_t_i in zip(src, dest_t)
            ]
            return np.array(err) < thresh

        model, inlier, _ = util.ransac(N_images, 3, model_f, err_f)
        A, Rt = model[0], model[1]

        Rt, src, dest = util.filter_lists([Rt, src, dest], inlier)

    # estimate using all inliers / all views
    A, Rt, k = _calibrate_camera_wrap(
        src, dest, image_width, image_height, n_dist_coeff
    )

    if ransac:
        return Rt, A, k, inlier
    else:
        return Rt, A, k


def warp_image(img, transform, x0, y0, xstep, ystep, w, h, method="linear"):
    xi, yi = np.meshgrid(np.arange(x0, x0 + w, xstep), np.arange(y0, y0 + h, ystep))
    samples = np.stack((xi.flatten(), yi.flatten()), axis=1)
    samples = transform(samples)
    v = scipy.interpolate.interpn(
        (
            np.arange(img.shape[0], dtype=np.float32),
            np.arange(img.shape[1], dtype=np.float32),
        ),
        img,
        np.fliplr(samples),
        method=method,
        bounds_error=False,
        fill_value=0,
    )
    return v.reshape((xi.shape[0], xi.shape[1]))


def warp_image_patches(img, transform, x0s, y0s, xstep, ystep, ws, hs, method="linear"):
    result = []
    for x0, y0, w, h in zip(x0s, y0s, ws, hs):
        result.append(warp_image(img, transform, x0, y0, xstep, ystep, w, h, method))
    return result


class Transform(ABC):
    """Base transform object"""

    @abstractmethod
    def __init__(self, src, dest, image_width=0, image_height=0, ransac=False):
        pass

    @abstractmethod
    def __call__(self, coords):
        pass

    @abstractmethod
    def inv(self):
        pass

    @abstractproperty
    def mask(self):
        pass

    @abstractproperty
    def reprojection_error(self):
        pass

    @abstractmethod
    def get_parameter_handles(self):
        pass

    @abstractmethod
    def update_parameters_by_handles(self):
        pass

    @abstractproperty
    def valid(self):
        pass

    def mean_scale(self):
        a = np.array([[0, 0], [10, 0], [10, 10], [0, 10]])
        at = self(a)
        s1 = (
            1 / 2 * (np.linalg.norm(at[1] - at[0]) + np.linalg.norm(at[3] - at[2])) / 10
        )
        s2 = (
            1 / 2 * (np.linalg.norm(at[2] - at[1]) + np.linalg.norm(at[0] - at[3])) / 10
        )
        s = 1 / 2 * (s1 + s2)
        d = np.round(s / 8)  # ensure that it is divisible by 8
        return float(d * 8)


class IdentityTransform(Transform):
    def __init__(self):
        pass

    def __call__(self, coords):
        return coords

    def inv(self):
        return self

    @property
    def mask(self):
        return None

    @property
    def reprojection_error(self):
        return 0.0

    def get_parameter_handles(self):
        return None

    def update_parameters_by_handles(self):
        return None

    def valid(self):
        return True


class MultiTransform(Transform):
    @abstractmethod
    def __getitem__(self, i):
        pass

    @abstractmethod
    def __len__(self):
        pass


class HomographyTransform(Transform):
    def __init__(
        self, src, dest, image_width=0, image_height=0, ransac=False, ransac_thres=0.1
    ):
        self._image_width = image_width
        self._image_height = image_height
        self._handles = None

        if ransac:
            # initial estimate to determine scale
            H = find_homography(src, dest)
            r = apply_homography(np.array([[0, 0], [1, 1]]), H)
            scale = np.linalg.norm(r[1] - r[0])

            H, self._mask, _ = find_homography_ransac(src, dest, ransac_thres * scale)
            src = src[self._mask]
            dest = dest[self._mask]
        else:
            self._mask = None

        self._H = find_homography(src, dest)
        self._err = np.mean(
            np.linalg.norm(apply_homography(src, self._H) - dest, axis=1)
        )

    def __call__(self, coords):
        return apply_homography(coords, self._H)

    def _get_sensitivity(self, axis):
        src = np.array([[0.0, 0.0]])
        t1 = self(src)
        src[0, axis] = 1.0
        t2 = self(src)
        return np.linalg.norm([t1 - t2])

    def inv(self):
        c = copy(self)
        c._H = np.linalg.pinv(self._H)
        c._H /= c._H[2, 2]
        return c

    @property
    def mask(self):
        return self._mask

    @property
    def reprojection_error(self):
        return self._err

    def get_parameter_handles(self):
        x_sens = self._get_sensitivity(0)
        y_sens = self._get_sensitivity(1)
        dx = 1 / x_sens * self._image_width
        dy = 1 / y_sens * self._image_width
        self._handles = np.array([[-dx, dy], [dx, dy], [dx, -dy], [-dx, -dy]])
        return self(self._handles).flatten()

    def update_parameters_by_handles(self, handles):
        handles = handles.reshape(self._handles.shape)
        self._H = find_homography(self._handles, handles)

    @property
    def valid(self):
        return self._H is not None


class FullTransform(Transform):
    def __init__(
        self, src, dest, image_width=0, image_height=0, ransac=False, n_dist_coeff=1
    ):
        Rt, self._A, self._dist = find_full_model(
            [src], [dest], n_dist_coeff, image_width, image_height, ransac=ransac
        )
        self._Rt = Rt[0]
        self._is_inverse = False
        self._success = True

    def __call__(self, coords):
        if self._success:
            return apply_pinhole(
                coords, self._Rt, self._A, self._dist, self._is_inverse
            )
        else:
            return None

    def inv(self):
        if self._success:
            c = copy(self)
            c._A = self._A
            c._Rt = self._Rt
            c._is_inverse = True
            return c
        else:
            return None

    @property
    def mask(self):
        return self._mask

    @property
    def reprojection_error(self):
        return None

    def get_parameter_handles(self, extrinsic=True, intrinsic=False, distortion=False):
        params = []

        if extrinsic:
            params.append(util.matrix2rodrigues(self._Rt[:, :3]))
            params.append(self._Rt[:, 3])
        if intrinsic:
            params.append(self._A[[0, 0, 1, 1], [0, 2, 1, 2]])
        if distortion:
            params.append(self._dist)

        return np.concatenate(params, axis=0)

    def update_parameters_by_handles(
        self, handles, extrinsic=True, intrinsic=False, distortion=False
    ):
        if extrinsic:
            self._Rt[:, :3] = util.rodrigues2matrix(handles[:3])
            self._Rt[:, 3] = handles[3:6]
            handles = handles[6:] if handles.shape[0] > 6 else None
        if intrinsic:
            self._A[[0, 0, 1, 1], [0, 2, 1, 2]] = handles[:4]
            handles = handles[4:] if handles.shape[0] > 4 else None
        if distortion:
            self._dist = handles

    @property
    def valid(self):
        return self._success


class FullMultiTransform(MultiTransform):
    def __init__(
        self,
        src,
        dest,
        image_width=0,
        image_height=0,
        ransac=False,
        n_dist_coeff=1,
        fixed_intrinsics=None,
    ):
        if ransac:
            assert fixed_intrinsics is None

        if ransac:
            self._Rt, self._A, self._dist, self._mask = find_full_model(
                src, dest, n_dist_coeff, image_width, image_height, ransac=ransac
            )
        else:
            if fixed_intrinsics is None:
                self._Rt, self._A, self._dist = find_full_model(
                    src, dest, n_dist_coeff, image_width, image_height, ransac=ransac
                )
            else:
                self._A = fixed_intrinsics[0]
                self._dist = fixed_intrinsics[1]
                self._Rt = _solve_PnP_wrap(src, dest, self._A, self._dist, n_dist_coeff)
        self._is_inverse = False
        self._success = True

    def __call__(self, coords):
        if self._success:
            return [
                apply_pinhole(x, rt, self._A, self._dist, self._is_inverse)
                for x, rt in zip(coords, self._Rt)
            ]
        else:
            return None

    def inv(self):
        if self._success:
            c = copy(self)
            c._A = self._A
            c._Rt = self._Rt
            c._is_inverse = True
            return c
        else:
            return None

    def mask(self):
        return self._mask

    def reprojection_error(self):
        return None

    def __getitem__(self, i):
        res = FullTransform.__new__(FullTransform)
        res._success = self._success
        res._A = self._A
        res._dist = self._dist
        res._Rt = self._Rt[i]
        res._err = None
        res._mask = None
        res._is_inverse = self._is_inverse
        return res

    def __len__(self):
        return len(self._Rt)

    def get_subset(self, indicator):
        res = FullMultiTransform.__new__(FullMultiTransform)
        res._success = self._success
        res._A = self._A
        res._dist = self._dist
        res._Rt = [x for i, x in enumerate(self._Rt) if indicator[i]]
        res._mask = None
        res._is_inverse = self._is_inverse
        return res

    def get_parameter_handles(
        self, items, extrinsic=True, intrinsic=False, distortion=False
    ):
        res = []
        res.append(
            self[items[0]].get_parameter_handles(
                extrinsic=extrinsic, intrinsic=intrinsic, distortion=distortion
            )
        )

        if extrinsic:
            for i in items[1:]:
                res.append(
                    self[i].get_parameter_handles(
                        extrinsic=extrinsic, intrinsic=False, distortion=False
                    )
                )

        return np.concatenate(res, axis=0)

    def update_parameters_by_handles(
        self, handles, items, extrinsic=True, intrinsic=False, distortion=False
    ):
        t = [self[i] for i in items]
        l = (
            t[0]
            .get_parameter_handles(
                extrinsic=extrinsic, intrinsic=intrinsic, distortion=distortion
            )
            .shape[0]
        )
        t[0].update_parameters_by_handles(
            handles[:l], extrinsic=extrinsic, intrinsic=intrinsic, distortion=distortion
        )
        handles = handles[l:] if handles.shape[0] > l else None
        self._A = t[0]._A
        self._dist = t[0]._dist
        self._Rt[items[0]] = t[0]._Rt

        if extrinsic and len(items) > 1:
            l = handles.shape[0] // (len(items) - 1)
            for ti, i in zip(t[1:], items[1:]):
                ti.update_parameters_by_handles(handles[:l])
                self._Rt[i] = ti._Rt
                handles = handles[l:] if handles.shape[0] > l else None

    @property
    def valid(self):
        return self._success



from pvinspect import data
from pvinspect.data.image import *
from pvinspect.data.image import _sequence
from pathlib import Path
from test.utilities import *
import pandas as pd


def test_sequence_element_access():
    seq = data.datasets.poly10x6(2)
    assert seq._images[0].path == seq[0].path
    assert seq._images[1].path == seq[1].path


def test_sequence_wrapper():
    @_sequence
    def _some_fn_mis(seq: ModuleImageSequence):
        assert type(seq) == ModuleImageSequence
        return seq

    @_sequence
    def _some_fn_is(seq: ImageSequence):
        assert type(seq) == ImageSequence
        return seq

    img = random_module_image()
    res = _some_fn_mis(img)
    assert isinstance(res, ModuleImage)

    img = random_image()
    res = _some_fn_is(img)
    assert isinstance(res, Image)

    img_seq = random_module_image_sequence()
    res = _some_fn_mis(img_seq)
    assert isinstance(res, ModuleImageSequence)

    img_seq = random_image_sequence()
    res = _some_fn_is(img_seq)
    assert isinstance(res, ImageSequence)


def test_sequence_wrapper_noarg():
    @_sequence()
    def _some_fn(seq: ModuleImageSequence):
        assert type(seq) == ModuleImageSequence
        return seq

    img = random_module_image()
    res = _some_fn(img)
    assert isinstance(res, ModuleImage)


def test_sequence_wrapper_nounwrap():
    @_sequence(True)
    def _some_fn(seq: ModuleImageSequence):
        assert type(seq) == ModuleImageSequence
        return seq

    img = random_module_image()
    res = _some_fn(img)
    assert isinstance(res, ModuleImageSequence)


def test_image_from_other():
    p = Path() / "other.png"
    img = random_image()
    img2 = Image.from_other(img, path=p)
    assert_equal(img._data, img2._data)
    assert img.modality == img2.modality
    assert img2.path == p.absolute()


def test_module_image_from_other():
    img = random_module_image()
    img2 = ModuleImage.from_other(img, cols=10, rows=6)
    assert_equal(img._data, img2._data)
    assert img.modality == img2.modality
    assert img.path == img2.path
    assert img2.cols == 10
    assert img2.rows == 6


def test_float_image_is_not_converted():
    data = np.array([[0.1]], dtype=np.float32)
    img = ModuleImage(data, EL_IMAGE, Path() / "test.png")
    assert img.dtype == DType.FLOAT
    data = np.array([[0.1]], dtype=np.float64)
    img = ModuleImage(data, EL_IMAGE, Path() / "test.png")
    assert img.dtype == DType.FLOAT


def test_apply_image_data():
    seq = random_image_sequence()
    original_data = seq[0].data.copy()

    def fn(x):
        x = x.copy()
        x[:] = 0.0
        return x

    seq2 = seq.apply_image_data(fn)
    assert_equal(seq2[0].data, 0.0)
    assert_equal(seq[0].data, original_data)


def test_image_as_type():
    img = random_image()
    assert img.dtype == DType.FLOAT
    img = img.as_type(DType.UNSIGNED_INT)
    assert img.dtype == DType.UNSIGNED_INT


def test_sequence_as_type():
    seq = random_image_sequence()
    assert seq.dtype == DType.FLOAT
    seq = seq.as_type(DType.UNSIGNED_INT)
    assert seq.dtype == DType.UNSIGNED_INT


def test_add_images():
    img1 = random_image()
    img2 = random_image()
    res = img1 + img2
    assert_equal(res.data, img1.data + img2.data)


def test_add_images_differing_dtypes():
    img1 = random_image()
    img2 = random_uint_image()
    try:
        res = img1 + img2
    except RuntimeError:
        return

    # expected an error
    assert False


def test_sub_images():
    img1 = random_image()
    img2 = random_image()
    res = img1 - img2
    assert_equal(res.data, img1.data - img2.data)


def test_sub_uint_images():
    img1 = random_uint_image()
    img2 = random_uint_image()
    res = img1 - img2
    expected = np.clip(
        img1.data.astype(np.int32) - img2.data.astype(np.int32), 0, 2 ** 16
    )
    assert_equal(res.data, expected)


def test_sub_images_differing_dtypes():
    img1 = random_image()
    img2 = random_uint_image()
    try:
        res = img1 - img2
    except RuntimeError:
        return

    # expected an error
    assert False


def test_mul_images():
    img1 = random_image()
    img2 = random_image()
    res = img1 * img2
    assert_equal(res.data, img1.data * img2.data)


def test_mul_images_differing_dtypes():
    img1 = random_image()
    img2 = random_uint_image()
    try:
        res = img1 * img2
    except RuntimeError:
        return

    # expected an error
    assert False


def test_truediv_images():
    img1 = random_image()
    img2 = random_image()
    res = img1 / img2
    assert_equal(res.data, img1.data / img2.data)


def test_truediv_images_nonfloat():
    img1 = random_image()
    img2 = random_uint_image()
    try:
        res = img1 / img2
    except RuntimeError:
        return

    # expected an error
    assert False


def test_floordiv_images():
    img1 = random_uint_image()
    img2 = random_uint_image()
    res = img1 // img2
    assert_equal(res.data, img1.data // img2.data)


def test_floordiv_images_differing_dtypes():
    img1 = random_image()
    img2 = random_uint_image()
    try:
        res = img1 // img2
    except RuntimeError:
        return

    # expected an error
    assert False


def test_mod_images():
    img1 = random_uint_image()
    img2 = random_uint_image()
    res = img1 % img2
    assert_equal(res.data, img1.data % img2.data)


def test_mod_images_differing_dtypes():
    img1 = random_image()
    img2 = random_uint_image()
    try:
        res = img1 % img2
    except RuntimeError:
        return

    # expected an error
    assert False


def test_pow_images():
    img1 = random_image()
    img2 = random_image()
    res = img1 ** img2
    assert_equal(res.data, img1.data ** img2.data)


def test_pow_images_differing_dtypes():
    img1 = random_image()
    img2 = random_uint_image()
    try:
        res = img1 ** img2
    except RuntimeError:
        return

    # expected an error
    assert False


def test_add_image_sequence():
    imgs1 = random_image_sequence()
    imgs2 = random_image_sequence()
    res = imgs1 + imgs2

    for img1, img2, r in zip(imgs1, imgs2, res):
        assert_equal(r.data, img1.data + img2.data)


def test_sub_image_sequence():
    imgs1 = random_image_sequence()
    imgs2 = random_image_sequence()
    res = imgs1 - imgs2

    for img1, img2, r in zip(imgs1, imgs2, res):
        assert_equal(r.data, img1.data - img2.data)


def test_mul_image_sequence():
    imgs1 = random_image_sequence()
    imgs2 = random_image_sequence()
    res = imgs1 * imgs2

    for img1, img2, r in zip(imgs1, imgs2, res):
        assert_equal(r.data, img1.data * img2.data)


def test_truediv_image_sequence():
    imgs1 = random_image_sequence()
    imgs2 = random_image_sequence()
    res = imgs1 / imgs2

    for img1, img2, r in zip(imgs1, imgs2, res):
        assert_equal(r.data, img1.data / img2.data)


def test_floordiv_image_sequence():
    imgs1 = random_image_sequence()
    imgs2 = random_image_sequence()
    res = imgs1 // imgs2

    for img1, img2, r in zip(imgs1, imgs2, res):
        assert_equal(r.data, img1.data // img2.data)


def test_mod_image_sequence():
    imgs1 = random_uint_image_sequence()
    imgs2 = random_uint_image_sequence()
    res = imgs1 % imgs2

    for img1, img2, r in zip(imgs1, imgs2, res):
        assert_equal(r.data, img1.data % img2.data)


def test_pow_image_sequence():
    imgs1 = random_image_sequence()
    imgs2 = random_image_sequence()
    res = imgs1 ** imgs2

    for img1, img2, r in zip(imgs1, imgs2, res):
        assert_equal(r.data, img1.data ** img2.data)


def test_image_meta():
    data = random_image().data
    path = random_image().path
    img_test = Image(data=data, path=path, meta={"key": "value"})

    assert img_test.has_meta("key")
    assert not img_test.has_meta("key2")
    assert img_test.get_meta("key") == "value"
    assert "key" in img_test.list_meta()

    # check from_other preserves meta
    img_other = Image.from_other(img_test, data=random_image().data)
    assert img_other.get_meta("key") == "value"

    # check from_other appends meta
    img_other2 = Image.from_other(img_other, meta={"key2": "value2"})
    assert img_other2.get_meta("key") == "value"
    assert img_other2.get_meta("key2") == "value2"
    assert "key" in img_other2.list_meta()
    assert "key2" in img_other2.list_meta()

    # but img_other not modified
    assert not img_other.has_meta("key2")


def test_image_meta_override():
    data = random_image().data
    path = random_image().path
    img_test = Image(data=data, path=path, meta={"key": "value"})
    img_other = Image.from_other(img_test, meta={"key": "othervalue"})
    assert img_other.get_meta("key") == "othervalue"


def test_image_data_is_immutable():
    data = random_image().data
    assert data.flags["WRITEABLE"] == False


def test_image_meta_is_immutable():
    img = random_image()
    arraymeta = np.zeros(1)
    othermeta = [1, 2, 3]
    img_with_meta = Image.from_other(
        img, meta={"arraymeta": arraymeta, "othermeta": othermeta}
    )

    # array should be immutable
    assert img_with_meta.get_meta("arraymeta").flags["WRITEABLE"] == False

    # other should return a copy
    assert img_with_meta.get_meta("othermeta") is not othermeta


def test_image_meta_to_pandas():
    img = random_image(meta={"k": 1})
    meta = img.meta_to_pandas()

    assert isinstance(meta, pd.Series)
    assert meta["k"] == 1


def test_image_meta_from_path():
    img = random_image(path=Path("img_d12.png"))
    img = img.meta_from_path(pattern=r"img_d(\d+)", key="k", target_type=int, group_n=1)
    meta = img.meta_to_pandas()

    assert meta["k"] == 12


def test_image_meta_from_path_transform():
    def t(x):
        assert isinstance(x, str)
        return int(x) * 2

    img = random_image(path=Path("img_d12.png"))
    img = img.meta_from_path(
        pattern=r"img_d(\d+)", key="k", target_type=int, group_n=1, transform=t
    )
    meta = img.meta_to_pandas()

    assert meta["k"] == 24


def test_sequence_meta_to_pandas():
    img1 = random_image(meta={"k": 1})
    img2 = random_image(meta={"k": 2})
    seq = ImageSequence([img1, img2], same_camera=False)
    meta = seq.meta_to_pandas()

    assert isinstance(meta, pd.DataFrame)
    assert meta.iloc[0]["k"] == 1
    assert meta.iloc[1]["k"] == 2


def test_sequence_meta_from_path():
    img1 = random_image(path=Path("img_d12.png"))
    img2 = random_image(path=Path("img_d13.png"))
    seq = ImageSequence([img1, img2], same_camera=False)
    seq = seq.meta_from_path(pattern=r"img_d(\d+)", key="k", target_type=int, group_n=1)
    meta = seq.meta_to_pandas()

    assert isinstance(meta, pd.DataFrame)
    assert meta.iloc[0]["k"] == 12
    assert meta.iloc[1]["k"] == 13


def test_sequence_pandas_method():
    img1 = random_image(meta={"k": 1})
    img2 = random_image(meta={"k": 2})
    seq = ImageSequence([img1, img2], same_camera=False)
    seq = seq.pandas.query("k == 1")

    assert len(seq) == 1
    assert seq[0].get_meta("k") == 1


def test_sequence_pandas_array_access_multiple():
    seq = random_image_sequence()
    sub = seq.pandas.iloc[[0, 1]]

    assert len(sub) == 2


def test_sequence_pandas_array_access_single():
    seq = random_image_sequence()
    img = seq.pandas.iloc[0]

    assert img == seq._images[0]


def test_from_other_drop_types():
    img1 = random_image(meta={"k1": 1, "k2": "x"})
    img2 = img1.from_other(img1, drop_meta_types=[str])

    assert img2.has_meta("k1")
    assert not img2.has_meta("k2")


def test_image_data_is_view():
    img = random_image()
    assert img.data.base is img._data


def test_image_data_is_readonly():
    img = random_image()
    assert img.data.flags["WRITEABLE"] == False


def test_lazy_load():
    class Loader:
        def __init__(self):
            self.loaded = False

        def __call__(self):
            self.loaded = True
            return np.ones((2), dtype=DTYPE_INT)

    loader = Loader()
    data = Image.LazyData(loader)
    img = Image(data=data, path=Path())

    assert loader.loaded == False
    assert np.sum(img.data) == 2
    assert loader.loaded == True


def test_lazy_load_checks():
    data = Image.LazyData(lambda: np.ones((2), dtype=np.uint8))
    img = Image(data=data, path=Path())

    assert img.data.dtype == DTYPE_UNSIGNED_BYTE


def test_load_cached():
    class Loader:
        def __init__(self):
            self.n_calls = 0

        def __call__(self):
            self.n_calls += 1
            return np.ones((2, 2))

    loader = Loader()
    data = Image.LazyData(loader)
    img = Image(data=data, path=Path("test.png"))

    data = img.data
    assert loader.n_calls == 1
    adata = img.data
    assert loader.n_calls == 1

    # make sure that we get the same instance
    assert data is adata


def test_lazy_from_other():
    class Loader:
        def __init__(self):
            self.n_calls = 0

        def __call__(self):
            self.n_calls += 1
            return np.ones((2, 2))

    loader = Loader()
    data = Image.LazyData(loader)
    img = Image(data=data, path=Path("test.png"))

    # create another image from img
    img2 = Image.from_other(img)

    # assure that image is not loaded
    assert loader.n_calls == 0



from test.utilities import assert_equal, random_ubyte_image_sequence
import torch as t
from pvinspect.integration.pytorch import *
from pvinspect.data import EL_IMAGE


def test_dataset():
    seq = random_ubyte_image_sequence(lazy=True, N=3)
    ds = Dataset(seq)

    assert len(ds) == 3
    for i in range(3):
        assert ds[i].data == seq[i].data


def test_dataset_with_meta():
    meta = [{"k1": v, "k2": -v} for v in range(3)]
    seq = random_ubyte_image_sequence(lazy=True, N=3, meta=meta)

    # only use a single meta attr
    ds = Dataset(seq, meta_attrs=["k1"])
    for i in range(3):
        x, y = ds[i]
        assert x.data == seq[i].data
        assert y == i

    # or both
    ds = Dataset(seq, meta_attrs=["k2", "k1"])
    for i in range(3):
        x, y, z = ds[i]
        assert x.data == seq[i].data
        assert y == -i
        assert z == i


def test_dataset_with_data_transform():
    seq = random_ubyte_image_sequence(lazy=True, N=3)
    t = lambda x: x + 1.0
    ds = Dataset(seq, data_transform=t)

    for i in range(3):
        assert_equal(ds[i].data, t(seq[i].data))


def test_dataset_with_meta_transform():
    meta = [{"k1": v, "k2": -v} for v in range(3)]
    seq = random_ubyte_image_sequence(lazy=True, N=3, meta=meta)
    t = lambda x: -x
    ds = Dataset(seq, meta_attrs=["k2", "k1"], meta_transforms={"k1": t})

    for i in range(3):
        _, y, z = ds[i]
        assert y == -i
        assert z == -i


def test_classification_dataset():
    meta = [{"c1": False, "c2": True}, {"c1": True, "c2": False}]
    seq = random_ubyte_image_sequence(lazy=True, N=2, meta=meta)
    ds = ClassificationDataset(seq, meta_classes=["c2", "c1"])

    for i in range(2):
        res = ds[i]
        x: t.Tensor = res[0]
        y: t.Tensor = res[1]
        assert isinstance(y, t.Tensor)
        assert len(res) == 2
        assert y.dtype == t.float32
        if i == 0:
            assert y.tolist() == [1.0, 0.0]
        else:
            assert y.tolist() == [0.0, 1.0]


def test_classification_dataset_additional_meta():
    meta = [{"c1": False, "c2": True, "k": 0}, {"c1": True, "c2": False, "k": 1}]
    seq = random_ubyte_image_sequence(lazy=True, N=2, meta=meta)
    ds = ClassificationDataset(seq, meta_classes=["c2", "c1"], meta_attrs=["k"])

    for i in range(2):
        res = ds[i]
        x: t.Tensor = res[0]
        y: t.Tensor = res[1]
        z: int = res[2]
        assert isinstance(y, t.Tensor)
        assert len(res) == 3
        assert y.dtype == t.float32
        if i == 0:
            assert y.tolist() == [1.0, 0.0]
            assert z == 0
        else:
            assert y.tolist() == [0.0, 1.0]
            assert z == 1


def test_classification_dataset_feed_result_bool():
    meta = [{"c1": False, "c2": True}, {"c1": True, "c2": False}]
    seq = random_ubyte_image_sequence(lazy=True, N=2, meta=meta)
    ds = ClassificationDataset(seq, meta_classes=["c2", "c1"])

    result = [
        t.tensor([False, True], dtype=t.bool),
        t.tensor([True, False], dtype=t.bool),
    ]
    seq2 = ds.result_sequence(result, prefix="xx_")
    assert seq2[0].get_meta("xx_c2") == False
    assert seq2[0].get_meta("xx_c1") == True
    assert seq2[1].get_meta("xx_c2") == True
    assert seq2[1].get_meta("xx_c1") == False


def test_classification_dataset_feed_result_float():
    meta = [{"c1": False, "c2": True}, {"c1": True, "c2": False}]
    seq = random_ubyte_image_sequence(lazy=True, N=2, meta=meta)
    ds = ClassificationDataset(seq, meta_classes=["c2", "c1"])

    result = [
        t.tensor([0.0, 1.0], dtype=t.float),
        t.tensor([1.0, 0.0], dtype=t.float),
    ]
    seq2 = ds.result_sequence(result, prefix="xx_")
    assert seq2[0].get_meta("xx_c2") == 0.0
    assert seq2[0].get_meta("xx_c1") == 1.0
    assert seq2[1].get_meta("xx_c2") == 1.0
    assert seq2[1].get_meta("xx_c1") == 0.0



from pvinspect import data
from pvinspect.data.io import ObjectAnnotations
from pvinspect.data.image import *


def test_poly10x6():
    seq = data.datasets.poly10x6()
    assert len(seq) == 20
    assert seq.modality == data.EL_IMAGE
    assert seq.shape == (2052, 2046)
    assert seq.same_camera == True


def test_caip_dataB():
    seq1, seq2, anns = data.datasets.caip_dataB()
    assert len(seq1) == 5
    assert seq1.modality == data.EL_IMAGE
    assert seq1.same_camera == False
    assert len(seq2) == 3
    assert seq2.modality == data.EL_IMAGE
    assert seq2.same_camera == False
    assert len(anns) == len(seq1) + len(seq2)


def test_caip_dataC():
    seq, anns = data.datasets.caip_dataC()
    assert len(seq) == 10
    assert seq.modality == data.EL_IMAGE
    assert seq.same_camera == True
    assert len(anns) == len(seq)


def test_caip_dataD():
    seq, anns = data.datasets.caip_dataD()
    assert len(seq) == 9
    assert seq.modality == data.EL_IMAGE
    assert seq.same_camera == True
    assert len(anns) == len(seq)


def test_calibration_ipv40CCD_FF():
    d = data.datasets.calibration_ipv40CCD_FF(N=2)
    assert isinstance(d, dict)
    assert "0A" in d.keys()
    assert len(d) == 3
    assert len(d["0A"]) == 2


def test_multi_module_detection():
    anns, imgs = data.datasets.multi_module_detection(N=2)
    assert isinstance(imgs, ImageSequence)
    for img in imgs:
        assert img.path.name in anns.keys()


def test_elpv():
    imgs = data.datasets.elpv(N=2)
    assert isinstance(imgs, ImageSequence)
    assert imgs.modality == EL_IMAGE

    # check meta available
    meta = [
        "defect_probability",
        "wafer",
        "crack",
        "inactive",
        "blob",
        "finger",
        "testset",
    ]

    for m in meta:
        assert imgs[0].has_meta(m)



from . import locate_corners
from . import locate_module
from pvinspect.common import transform
from .config import (
    CORNER_DETECTION_PATCH_SIZE,
    RANSAC_THRES,
    PREPROC_LOWER_PERCENTILE,
    PREPROC_UPPER_PERCENTILE,
)
import numpy as np
from typing import Tuple
from skimage import exposure
from scipy.ndimage.filters import median_filter
from matplotlib import pyplot as plt


def apply(
    img: np.ndarray,
    n_cols: int,
    m_rows: int,
    is_module_detail: bool,
    orientation: str = None,
) -> Tuple[transform.Transform, np.ndarray, np.ndarray, np.ndarray]:
    """Locate the module in given EL image

    Args:
        img (np.array): The image
        n_cols (int): Number of columns of cells
        m_rows (int): Number of rows of cells
        is_module_detail (bool): Is this image showing only a part of a module?
        orientation (std): Orientation of the module (default: None = determine automatically)

    Returns:
        result: Returns a tuple (transform, model corners, detected corners, accepted flags)
    """

    # preprocessing
    img = img.copy()
    x = np.percentile(img.flatten(), PREPROC_UPPER_PERCENTILE, interpolation="lower")
    img[img > x] = x
    x = np.percentile(img.flatten(), PREPROC_LOWER_PERCENTILE, interpolation="lower")
    img[img < x] = x
    img = exposure.rescale_intensity(img)
    img = img / np.max(img)

    # locate module
    module_bb_fit = locate_module.locate_module(img, n_cols, m_rows)
    if module_bb_fit is None:
        return None, None, None, None  # failed
    module_bb = locate_module.module_boundingbox_model(
        module_bb_fit, n_cols, m_rows, orientation
    )
    locate_transform = transform.HomographyTransform(module_bb, module_bb_fit)

    # fit outer corners
    outer_corners = locate_corners.outer_corners_model(n_cols, m_rows)
    outer_corners_fit, outer_accepted = locate_corners.fit_outer_corners(
        img,
        outer_corners,
        locate_transform,
        CORNER_DETECTION_PATCH_SIZE,
        n_cols,
        m_rows,
        is_module_detail=is_module_detail,
    )
    outer_corners_fit = locate_transform(outer_corners_fit)
    # plt.imshow(img, cmap='gray')
    # plt.scatter(outer_corners_fit[outer_accepted, 0], outer_corners_fit[outer_accepted, 1], s=1, c='red')
    # plt.show()

    # reestimate transform
    locate_transform = transform.HomographyTransform(
        outer_corners[outer_accepted], outer_corners_fit[outer_accepted], ransac=True
    )

    # fit inner corners
    # TODO:
    #   - fix inaccurate detection of x/y location if considered edges differ in length
    #   - refine location of module corners first
    inner_corners = locate_corners.inner_corners_model(n_cols, m_rows)
    inner_corners_fit, inner_accepted = locate_corners.fit_inner_corners(
        img, inner_corners, locate_transform, CORNER_DETECTION_PATCH_SIZE
    )
    inner_corners_fit = locate_transform(inner_corners_fit)
    locate_transform = transform.HomographyTransform(
        np.concatenate((module_bb, inner_corners[inner_accepted]), axis=0),
        np.concatenate((module_bb_fit, inner_corners_fit[inner_accepted]), axis=0),
    )
    # plt.imshow(img, cmap='gray')
    # plt.scatter(inner_corners_fit[inner_accepted, 0], inner_corners_fit[inner_accepted, 1], s=1, c='red')
    # plt.show()

    # reestimate transform using all correspondences
    all_corners_model = np.concatenate((inner_corners, outer_corners), axis=0)
    all_corners_fit = np.concatenate((inner_corners_fit, outer_corners_fit), axis=0)
    all_accepted = np.concatenate((inner_accepted, outer_accepted), axis=0)
    locate_transform = transform.HomographyTransform(
        all_corners_model[all_accepted],
        all_corners_fit[all_accepted],
        img.shape[0],
        img.shape[1],
        ransac=True,
        ransac_thres=RANSAC_THRES,
    )
    all_accepted[all_accepted] = np.logical_and(
        all_accepted[all_accepted], locate_transform.mask
    )

    return locate_transform, all_corners_model, all_corners_fit, all_accepted



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pvinspect",
    version="0.4.1",
    author="Mathis Hoffmann",
    author_email="mathis.hoffmann@fau.de",
    description="Provides methods for the analysis of PV modules using different modalities",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ma0ho/pvinspect",
    packages=setuptools.find_packages(),
    python_requires=">=3.7, <3.9",
    install_requires=[
        "numpy >= 1.16.5",
        "scikit-image >= 0.16",
        "matplotlib",
        "pathlib",
        "tqdm",
        "pytest",
        "scikit-image",
        "scipy",
        "opencv-python",
        "requests",
        "googledrivedownloader",
        "opencv-python",
        "shapely",
        "optuna",
        "sklearn",
        "pre-commit",
        "pillow >= 7.2.0",
        "pandas >= 1.0.0",
        "onnxruntime",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
)



"""Flatfield and lens calibration"""
from pvinspect.data.image import *
from pvinspect.data.image import _sequence
from pvinspect.common.exceptions import (
    MisconfigurationException,
    InvalidArgumentException,
)
from pvinspect.common.types import PathOrStr
from typing import List, Tuple, Union
import numpy as np
from skimage import img_as_float
from tqdm.autonotebook import trange, tqdm
import numpy as np
import cv2
from skimage.exposure import rescale_intensity
import os
import logging
import pickle
from shapely.geometry import Polygon
from shapely import affinity
from skimage import transform, morphology, measure, filters
from typing import Union
import logging


def _calibrate_flatfield(
    images: List[np.ndarray], targets: List[float], order: int
) -> np.array:

    imgs = [img.flatten() for img in images]

    order = min(len(imgs) - 1, 2) if order == -1 else order
    coeff = np.empty((order + 1, imgs[0].shape[0]), dtype=images[0].dtype)

    if order == 1 and len(images) == 2:
        coeff[1] = (targets[1] - targets[0]) / (imgs[1] - imgs[0])
        coeff[0] = targets[0] - coeff[1] * imgs[0]
    else:
        for i in trange(imgs[0].shape[0]):
            coeff[:, i] = np.polynomial.polynomial.polyfit(
                [img[i] for img in imgs], targets, order
            )

    # perserve mean of ff-image with highest intensity
    maxi = np.argmax(targets)
    mean = np.mean(images[maxi])
    coeff *= mean

    return coeff.reshape((order + 1, images[0].shape[0], images[0].shape[1]))


def calibrate_flatfield(
    images: Union[ImageSequence, List[ImageSequence]],
    targets: List[float],
    order: int = -1,
    use_median: bool = True,
) -> np.array:
    """Perform flat-field calibration. Note that there might be several calibration shots
    with the same normalization target. In that case, a least-squares estimate is computed.

    Args:
        images (Union[ImageSequence, List[ImageSequence]]): Sequence of calibration shots or list of sequences
        targets (List[float]): Corresponding list of calibration targets. If images is an ImageSequence, specify
            one target per element of the sequence. If images is a list of ImageSequence, specify one target
            per element of the list. The targets specify normalized intensity of the flatfield calibration image.
        order (int): Order of compensation polynomial. The order is chosen automatically, if order == -1 (default)
        use_median (bool): Use the median of all images with the same target

    Returns:
        coeff: order+1 coefficients starting with the lowest order coefficient
    """
    if order == 0:
        raise InvalidArgumentException("Order of flat-field-polynom must not be 0")
    if order > len(images) - 1:
        raise MisconfigurationException(
            "Need at least {:d} images to calibrate a flat-field-polynom of degree {:d} ({:d} given)".format(
                order + 1, order, len(images)
            )
        )

    # flatten multiple sequences
    if isinstance(images, list):
        images_new, targets_new = list(), list()
        for l, t in zip(images, targets):
            targets_new += [t] * len(l)
            images_new += l.images
        targets = targets_new
        images = ImageSequence(images=images_new, same_camera=True)

    # assure float type
    images = images.as_type(DType.FLOAT)

    if use_median:
        images_new = list()
        targets_new = list()

        for i1, t1 in enumerate(set(targets)):
            group = list()
            for i2, t2 in enumerate(targets):
                if t1 == t2:
                    group.append(images[i2].data)
            images_new.append(np.median(group, axis=0))
            targets_new.append(t1)

        return _calibrate_flatfield(images=images_new, targets=targets_new, order=order)

    return _calibrate_flatfield(
        images=[img.data for img in images], targets=targets, order=order
    )


def calibrate_distortion(
    images: ImageSequence, checkerboard_size: Tuple[int, int]
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, Tuple[int, int, int, int]]:
    """Perform lens calibration

    Args:
        images (ImageSequence): Sequence of calibration shots
        checkerboard_size (Tuple[int, int]): Size of the checkerboard (outer corners do not count)

    Returns:
        mtx: Matrix of intrinsic camera parameters
        dist: Vectors of distortion coefficients (k1, k2, p1, p2, k3)
        newcameramtx: Matrix that performs additional scaling to account for black borders
        roi: ROI of valid pixels
    """
    criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 30, 0.001)

    # prepare object points, like (0,0,0), (1,0,0), (2,0,0) ....,(6,5,0)
    objp = np.zeros((checkerboard_size[0] * checkerboard_size[1], 3), np.float32)
    objp[:, :2] = np.mgrid[
        0 : checkerboard_size[0], 0 : checkerboard_size[1]
    ].T.reshape(-1, 2)

    # Arrays to store object points and image points from all the images.
    objpoints = []  # 3d point in real world space
    imgpoints = []  # 2d points in image plane.

    for img in tqdm(images):
        p = np.percentile(img.data, [5, 95])
        img = rescale_intensity(
            img.data, in_range=tuple(p.tolist()), out_range=(0, 255)
        ).astype(np.uint8)

        # Find the chess board corners
        ret, corners = cv2.findChessboardCornersSB(img, checkerboard_size)

        if ret == True:
            objpoints.append(objp)
            imgpoints.append(corners)

    # sanity check
    if len(objpoints) < 0.8 * len(images):
        logging.warn(
            "Checkerboard detection failed on more than 20% percent of the images (only {:d} succeeded). Please consider preprocessing the images, such that the checkerboard has a better contrast.".format(
                len(objpoints)
            )
        )

    ret, mtx, dist, rvecs, tvecs = cv2.calibrateCamera(
        objpoints, imgpoints, images[0].shape[::-1], None, None
    )
    newcameramtx, roi = cv2.getOptimalNewCameraMatrix(
        mtx, dist, img.shape, 1, img.shape
    )

    return mtx, dist, newcameramtx, roi


@_sequence
def _compensate_flatfield(
    sequence: ModuleImageOrSequence, coeff: np.ndarray
) -> ModuleImageOrSequence:
    """Low level method to perform flat field correction"""

    def fn(data, coeff):
        res = coeff[0].copy()
        data = data.copy()
        for i in range(1, coeff.shape[0]):
            res += coeff[i] * data
            data *= data

        if res.min() < 0.0 or res.max() > 1.0:
            logging.warn(
                "Image exceeds datatype limits after FF-compensation. Clipping to limits.."
            )
            res = np.clip(res, 0.0, 1.0)

        return res

    return sequence.apply_image_data(fn, coeff)


def _do_locate_reference_cell(
    image: Image, reference_area: int, scale: float = 0.1,
) -> Polygon:
    """Perform localization of reference cell"""

    # filter + binarize
    image_f = transform.rescale(image.data, scale)

    thresh = filters.threshold_multiotsu(image_f, classes=5)[0]
    image_t = image_f > thresh

    # find regions
    labeled = morphology.label(image_t)
    regions = measure.regionprops(labeled)

    # drop areas that are not approximately square
    regions = [
        r
        for r in regions
        if np.abs(r.bbox[2] - r.bbox[0]) / np.abs(r.bbox[3] - r.bbox[1]) > 0.8
        and np.abs(r.bbox[2] - r.bbox[0]) / np.abs(r.bbox[3] - r.bbox[1]) < 1.8
    ]

    # convert to Polygon
    tmp: List[Polygon] = []
    for r in regions:
        bbox = [int(r.bbox[i] * 1 / scale) for i in range(4)]
        y0, x0 = max(0, bbox[0]), max(0, bbox[1])
        y1, x1 = (
            min(image.shape[0], bbox[2]),
            min(image.shape[1], bbox[3]),
        )
        box = Polygon.from_bounds(x0, y0, x1, y1)
        tmp.append(box)

    # drop boxes that intersect with others
    regions = [
        r for r in tmp if not np.any([x.intersects(r) and x is not r for x in tmp])
    ]

    if len(regions) == 0:
        return None

    # process regions
    area_dev = [np.abs((r.area - reference_area) / reference_area) for r in regions]
    min_dev_idx = np.argmin(area_dev)

    if area_dev[min_dev_idx] < 2.0:
        return affinity.scale(regions[min_dev_idx], xfact=0.5, yfact=0.5)
    else:
        return None


def _do_reference_scaling(
    image: Image, area: Polygon, ref: Union[float, int],
) -> Image:
    """Compute mean intensity over area area and rescale image"""
    original_type = image.dtype

    # compute statistics and scale
    x, y = list(zip(*area.exterior.coords))
    xmin, xmax = np.min(x), np.max(x)
    ymin, ymax = np.min(y), np.max(y)
    median = np.median(image.data[int(ymin) : int(ymax), int(xmin) : int(xmax)])
    data = (image.data * (ref / median)).astype(DTYPE_FLOAT)

    # convert to original datatype
    if original_type == DType.INT:
        data = data.astype(DTYPE_INT)
    elif original_type == DType.UNSIGNED_INT:
        data = data.astype(DTYPE_UNSIGNED_INT)

    return image.from_other(image, data=data, meta={"calibration_reference_box": area})


@_sequence
def compensate_flatfield(
    sequence: ModuleImageOrSequence, coeff: np.ndarray
) -> ModuleImageOrSequence:
    """Perform flat field correction

    Args:
        sequence (ModuleImageOrSequence): Sequence of images or single image
        coeff (np.ndarray): Compensation coefficients

    Returns:
        sequence: The corrected images
    """
    otype = sequence.dtype
    sequence = sequence.as_type(DType.FLOAT)
    sequence = _compensate_flatfield(sequence, coeff)
    if otype is not None:
        sequence = sequence.as_type(otype)

    return sequence


@_sequence
def compensate_distortion(
    sequence: ModuleImageOrSequence,
    mtx: np.ndarray,
    dist: np.ndarray,
    newcameramtx: np.ndarray,
    roi: Tuple[int, int, int, int],
) -> ModuleImageOrSequence:
    """Perform lens distortion correction

    Args:
        sequence (ModuleImageOrSequence): Sequence of images or single image
        mtx (np.ndarray): Matrix of instrinsic camera parameters
        dist (np.ndarray): Matrix of distortion coefficients
        newcameramtx (np.ndarray): Matrix that performs additional scaling to account for black borders
        roi (Tuple[int, int, int, int]): ROI of valid pixels

    Returns:
        sequence: The corrected images
    """

    def corr(x):
        dst = cv2.undistort(x, mtx, dist, None, newcameramtx)
        return dst[roi[0] : roi[2], roi[1] : roi[3]]

    return sequence.apply_image_data(corr)


class Calibration:
    """Handles camera calibration and processing of images"""

    def __init__(self, ff_poly_order: int = -1, ff_use_median: bool = True):
        """Initialize calibration object

        Args:
            ff_poly_order (int): Order of flat-field compensation polynomial. The order is chosen automatically, if order == -1 (default)
            ff_use_median (bool): Use the median of all images with the same target
        """

        self._ff_poly_order = ff_poly_order
        self._ff_use_median = ff_use_median
        self._ff_calibration = None
        self._ff_dtype = None
        self._dist_calibration = None

    def calibrate_flatfield(
        self, images: Union[ImageSequence, List[ImageSequence]], targets: List[float]
    ):
        """Perform flat-field calibration. Note that there might be several calibration shots
        with the same normalization target. In that case, a least-squares estimate is computed.

        Args:
            images (Union[ImageSequence, List[ImageSequence]]): Sequence of calibration shots
            targets (List[float]): Corresponding list of calibration targets. If images is an ImageSequence, specify
                one target per element of the sequence. If images is a list of ImageSequence, specify one target
                per element of the list. The targets specify normalized intensity of the flatfield calibration image.
        """
        if np.max(targets) != 1.0:
            raise RuntimeError("One of the calibration targets must equal 1.0")
        if np.min(targets) != 0.0:
            raise RuntimeError("One of the calibration targets must equal 0.0")

        # store dtype for later checks
        self._ff_dtype = images[0].dtype

        self._ff_calibration = calibrate_flatfield(
            images=images,
            targets=targets,
            order=self._ff_poly_order,
            use_median=self._ff_use_median,
        )

    def calibrate_distortion(
        self, images: ImageSequence, checkerboard_size: Tuple[int, int]
    ):
        """Perform lens calibration

        Args:
            images (ImageSequence): Sequence of calibration shots
            checkerboard_size (Tuple[int, int]): Size of the checkerboard (outer corners do not count)
        """

        self._dist_calibration = calibrate_distortion(
            images=images, checkerboard_size=checkerboard_size
        )

    def process(
        self,
        images: ImageOrSequence,
        flatfield: bool = True,
        distortion: bool = True,
        reference_cell_area: int = None,
        reference_intensity_key: str = None,
    ):
        """Process images and compensate camera artifacts, optionally normalize using reference cell

        Args:
            images (ImageOrSequence): Sequence of images or single image
            clip_result (bool): Clip the result such that all pixels are in the range [0,1]
            flatfield (bool): Perform flat-field compensation
            distortion (bool): Perform distortion compensation
            reference_cell_area (int): Approximate area of the reference cell in pixels. If set
                images are automatically normalizes such that the reference cell intensity
                matches values given by meta attribute reference_intensity_key
            reference_intensity_key (str): Meta key that holds the reference intensity
        """

        if self._ff_calibration is not None and flatfield:
            if images.dtype != self._ff_dtype:
                logging.warn(
                    "Datatype of images ({}) differs from calibration images {}. This might lead to incorrect results.".format(
                        images.dtype, self._ff_dtype
                    )
                )
            logging.info("Processing flatfield compensation..")
            images = compensate_flatfield(images, self._ff_calibration)
        elif flatfield:
            logging.warn(
                "No flat-field calibration data available. Use calibrate_flatfield to perform calibration. Skipping flat-field compensation.."
            )

        if self._dist_calibration is not None and distortion:
            logging.info("Processing distortion compensation..")
            images = compensate_distortion(images, *self._dist_calibration)
        elif distortion:
            logging.warn(
                "No distortion calibration data available. Use calibrate_distortion to perform calibration. Skipping distortion compensation.."
            )

        if reference_cell_area is not None and reference_intensity_key is not None:

            def fn(x: Image):
                box = _do_locate_reference_cell(x, reference_cell_area)
                if box is not None:
                    return _do_reference_scaling(
                        x, box, x.get_meta(reference_intensity_key)
                    )
                else:
                    logging.warn(
                        "Reference cell could not be detected for {}. Image is left unscaled!".format(
                            x.path.name
                        )
                    )
                    return x

            logging.info("Processing reference scaling..")
            images = images.apply(fn)

        return images

    def save(self, path: PathOrStr):
        """Save calibration data to disk

        Args:
            path (PathOrStr): Target to save file to
        """

        with open(path, "wb") as f:
            pickle.dump(self, f)


def load_calibration(path: PathOrStr) -> Calibration:
    """Load calibration data from file

    Args:
        path (PathOrStr): Path to calibration file
    
    Returns:
        calibration (Calibration): The calibration object
    """

    with open(path, "rb") as f:
        return pickle.load(f)



import logging

from numpy.core.arrayprint import BoolFormat
from pvinspect.data import ImageSequence, Image, DType
from typing import Tuple, List, Optional, Callable, Any, Dict, Union, TypeVar
import numpy as np

try:
    import torch as t
    from torch.utils.data import Dataset as TorchDataset  # need this for mypy
except ImportError as e:
    logging.error("You need to install PyTorch in order to use the PyTorch-integration")
    raise e


# return type of data transform
D = TypeVar("D")


class Dataset(TorchDataset):
    def __init__(
        self,
        data: ImageSequence,
        data_transform: Optional[Callable[[np.ndarray], D]] = None,
        meta_attrs: List[str] = list(),
        meta_transforms: Dict[str, Callable[[Any], Any]] = dict(),
    ):
        """A PyTorch-compatible Dataset implementation for ImageSequence

        Args:
            data (ImageSequence): The ImageSequence that should be wrapped in the dataset
            data_transform (Optional[Callable[[np.ndarray], D]]): Callable that transforms data, use 
                this for online data augmentation/conversion etc.
            meta_attrs (List[str]): Meta attributes that should be returned in addition to the data
                on accessing single elements of the dataset
            meta_transforms (Dict[str, Callable[[Any], Any]]): Use this to specify additional transforms
                for individual meta attributes
        """
        super().__init__()
        self._data = data
        self._meta_attrs = meta_attrs

        # make sure that data_transform is callable
        identity = lambda x: x
        self._data_transform = (
            data_transform if data_transform is not None else identity
        )

        # make sure that for every meta attribute there is a callable transform
        n_meta_transforms = len(meta_transforms) if meta_transforms is not None else 0
        if meta_attrs is not None:
            meta_transforms_keys = (
                meta_transforms.keys() if meta_transforms is not None else []
            )
            self._meta_transforms = [
                meta_transforms[k] if k in meta_transforms_keys else identity
                for k in meta_attrs
            ]
        else:
            self._meta_transforms = []

        if not data[0].lazy:
            logging.warn(
                "The ImageSequence used to construct this Dataset is not lazy loaded. "
                "We recommend using lazy loaded data in combination with PyTorch."
            )

        if not data.dtype == DType.UNSIGNED_BYTE and not data.dtype == DType.FLOAT:
            logging.warn(
                "Image datatype is incompatible to PyTorch. "
                "Please consider converting them in advance"
            )

    def __getitem__(self, index: int) -> Union[D, Tuple[Any, ...]]:
        # PyTorch does not like uint>8 -> perform automatic conversion
        array = self._data[index].data
        if array.dtype == np.float64:
            array = array.astype(np.float32)

        # image data
        d = self._data_transform(array)

        # meta -> labels
        if self._meta_attrs is not None:
            meta = [
                self._meta_transforms[i](self._data[index].get_meta(k))
                for i, k in enumerate(self._meta_attrs)
            ]
        else:
            meta = []

        ret = tuple([d] + meta)
        if len(ret) > 1:
            return ret
        else:
            return ret[0]

    def __len__(self) -> int:
        return len(self._data)


class ClassificationDataset(Dataset):

    ClassificationMeta = Dict[str, Union[float, bool]]

    def _meta_to_tensor(self, meta: List[Any]) -> t.Tensor:
        return t.tensor([meta[i] for i in self._meta_classes_idx], dtype=t.float32)

    def _tensor_to_meta_dict(self, tensor: t.Tensor, prefix: str) -> ClassificationMeta:
        l = tensor.tolist()
        return {prefix + k: v for k, v in zip(self._meta_classes, l)}

    def __init__(
        self,
        data: ImageSequence,
        meta_classes: List[str],
        data_transform: Optional[Callable[[np.ndarray], D]] = None,
        meta_attrs: List[str] = list(),
        meta_transforms: Dict[str, Callable[[Any], Any]] = dict(),
    ):
        """A PyTorch-compatible Dataset implementation for ImageSequence

        Args:
            data (ImageSequence): The ImageSequence that should be wrapped in the dataset
            meta_classes (List[str]): List of meta attributes that should become part of the
                one-hot encoded target vector. These meta attributes must be boolean.
            data_transform (Optional[Callable[[np.ndarray], D]]): Callable that transforms data, use 
                this for online data augmentation/conversion etc.
            meta_attrs (List[str]): Meta attributes that should be returned in addition to the data
                and the one-hot encoded target on accessing single elements of the dataset
            meta_transforms (Dict[str, Callable[[Any], Any]]): Use this to specify additional transforms
                for individual meta attributes. These apply to members of meta_classes as well
        """
        meta_attrs = meta_classes + meta_attrs
        super().__init__(
            data=data,
            data_transform=data_transform,
            meta_attrs=meta_attrs,
            meta_transforms=meta_transforms,
        )
        self._meta_classes = meta_classes
        self._meta_classes_idx = [self._meta_attrs.index(k) for k in meta_classes]

    def __getitem__(self, index: int) -> Union[D, Tuple[D, t.Tensor], Tuple[Any, ...]]:
        res = list(super().__getitem__(index))
        x = res[0]
        y = self._meta_to_tensor(res[1:])
        if len(res) > len(self._meta_classes) + 1:
            other = res[len(self._meta_classes) + 1 :]
        else:
            other = []
        res = [x, y] + other
        return tuple(res)

    def result_sequence(self, results: List[t.Tensor], prefix: str) -> ImageSequence:
        """Feed classification results back to obtain an ImageSequence with predictions

        Args:
            results (List[torch.Tensor]): List of predictions from the network. Convert to
                boolean or float before calling this. The order of elements must be the same as in
                the underlying ImageSequence.
            prefix (str): Prefix that is appended to the corresponding meta keys
        
        Returns:
            result (ImageSequence): The ImageSequence with additional meta data
        """
        assert len(results) == len(self._data)
        assert results[0].dtype == t.float or results[0].dtype == t.bool

        # build an inverse map from Image to index
        imap = {k: v for v, k in enumerate(self._data.images)}

        # assign meta data
        def meta_fn(x: Image) -> ClassificationDataset.ClassificationMeta:
            idx = imap[x]
            return self._tensor_to_meta_dict(results[idx], prefix=prefix)

        return self._data.meta_from_fn(meta_fn)



class UnsupportedModalityException(BaseException):
    """This exception indicates that the given imaging modality is not supported"""

    def __init__(self, name):
        super().__init__('Modality "{}" is not supported'.format(name))
